"""Main interface for generating and visualizing an object's type tree.

- Use :class:`Tree` to generate the type tree as an object, which can be
  traversed as a subclass of a nested tuple.
- Use :func:`print_tree` to directly print the tree view.
- Use :func:`view_tree` to open the tree view as an interactive GUI.
"""

from .typetree import *

__version__ = '0.1.6'

__all__ = [
    'Tree',
    'Subtree',
    'print_tree',
    'view_tree',
    'Template',
    'DOM',
    'HTML',
    'XML',
    'KeyType'
]



# -*- coding: utf-8 -*-
"""Object type tree analyzer. The GUI is handled by a separate module."""

import dataclasses
import enum
import functools
import json
import re

from collections.abc import Callable
from typing import Any, Type

try:
    from viewer import tree_viewer
except (ModuleNotFoundError, ImportError):
    from .viewer import tree_viewer

_DEFAULT_MAX_LINES = 1000
_DEFAULT_MAX_SEARCH = 100_000
_DEFAULT_MAX_DEPTH = 20
_DEFAULT_MAX_BRANCHES = float('inf')

# Pre-compiled format for _KeyType.INDEX and _KeyType.SLICE
_RANGE_REGEX = re.compile(r'^\[(\d+)(?::(\d+))?]$')


@functools.total_ordering
class KeyType(enum.Enum):
    """Node key types.

    Each key will be displayed differently based on their type.
    """

    NONE = 0, 'None'
    """For the root node, which has no key."""

    ATTR = 1, 'Attribute'
    """For object attributes (starts with a dot).

    Examples: .attr, ._protected, .__special__.
    """

    MAP = 2, 'Mapping'
    """For dict-like Mapping keys.

    Examples: ['key'], [datetime.date(1970, 1, 1)].
    """

    INDEX = 3, 'Sequence'
    """For array-like Sequence indices or slices.

    The indices must be consecutive integers.
    Examples: [2], [4:7], [:3], [:].
    """

    SET = 4, 'Collection'
    """For set-like Collections, which have item counters, but no key.

    Example: (×3).
    The path to the item will be set to .copy().pop() for convenience,
    even though not all Collections accept these methods.
    """

    @classmethod
    def path(cls, key_type: 'KeyType', value: Any = None) -> str:
        """Key path from the parent object to the current one."""
        match key_type, value:
            case cls.NONE, None:
                return ''
            case cls.NONE, str(x):
                return x
            case cls.ATTR, _:
                return f'.{value!s}'
            case cls.MAP, _:
                return f'[{value!r}]'
            case cls.SET, int():
                return '.copy().pop()'
            case cls.INDEX, int(x):
                return f'[{x}]'
            case cls.INDEX, None:
                return '[:]'
            case cls.INDEX, (int(x), int(y)):
                if x + 1 == y:
                    return f'[{x}]'
                if x == 0:
                    return f'[:{y}]'
                return f'[{x}:{y}]'
        raise TypeError(f"Invalid key type '{key_type}' or value '{value}'")

    @classmethod
    def str(cls, key_type: 'KeyType', value: Any = None) -> str:
        """Key label to be displayed in the tree view."""
        match key_type, value:
            case cls.NONE, None:
                return ''
            case cls.SET, int(x):
                return '(×{:d}) '.format(x)
            case _:
                return f'{cls.path(key_type, value)}: '

    def __lt__(self, other: 'KeyType'):
        """For sorting. Respect the declaration order."""
        if not isinstance(other, type(self)):
            return TypeError
        return self.value < other.value


@functools.total_ordering
class _NodeKey:
    """Node key for string representation and for sorting."""

    def __init__(self, key_type: KeyType, value: Any = None):
        self._str: str = KeyType.str(key_type, value)
        self._path: str = KeyType.path(key_type, value)
        self._type: KeyType = key_type
        self._counter: int = 1
        if key_type == KeyType.SET:
            if not isinstance(value, int):
                raise TypeError
            self._counter = value
        self._slice: tuple[int, int] | None = None
        if key_type == KeyType.INDEX:
            if isinstance(value, int):
                self._slice = value, value + 1
            else:
                self._slice = value
        # Hash is unique for Sets
        self._hash: int = hash((
            type(self),
            self._type,
            self._slice,
            id(self)*(self._type == KeyType.SET)
        ))

    @property
    def path(self) -> str:
        return self._path

    @property
    def type(self) -> KeyType:
        return self._type

    @property
    def counter(self) -> int:
        return self._counter

    @property
    def slice(self) -> tuple[int, int] | None:
        return self._slice

    def reset_counter(self):
        self._counter = 1
        self._str = KeyType.str(self.type, self._counter)

    def increment_counter(self):
        self._counter += 1
        self._str = KeyType.str(self.type, self._counter)

    def __str__(self) -> str:
        return self._str

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            raise NotImplementedError
        if self._type == other._type == KeyType.SET:
            return False
        return self._str == other._str

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            raise NotImplementedError
        if self._str == other._str:
            return False
        if self._slice is not None:
            if other._slice is None:
                return self._type < other._type
            return self._slice < other._slice
        if self._type == other._type:
            if self._type == KeyType.SET:
                if self._counter != other._counter:
                    return self._counter < other._counter
                return self._hash < other._hash
            return self._str < other._str
        return self._type < other._type

    def __hash__(self) -> int:
        return self._hash


def get_itself(var: Any) -> Any:
    """Do nothing, return itself."""
    return var


def get_type_name(var: Any) -> str:
    """Return the input's type name."""
    return str(type(var).__name__)


def getattr_maker(attr: str) -> Callable[[Any], Any]:
    """Build a partial `getattr` for a fixed attribute."""
    def attr_get(var: Any, attr_copy: str = attr) -> Any:
        return getattr(var, attr_copy)
    return attr_get


@dataclasses.dataclass(slots=True, frozen=True)
class Template:
    """Default template for configuration properties of :class:`Tree`."""

    items_lookup: Callable[[Any], Any] = get_itself
    type_name_lookup: Callable[[Any], str] = get_type_name
    value_lookup: Callable[[Any], Any] = get_itself
    sort_keys: bool = True
    show_lengths: bool = True
    include_attributes: bool = True
    include_dir: bool = False
    include_protected: bool = False
    include_special: bool = False
    max_search: float = _DEFAULT_MAX_SEARCH
    max_depth: float = _DEFAULT_MAX_DEPTH
    max_branches: float = _DEFAULT_MAX_BRANCHES


@dataclasses.dataclass(slots=True, frozen=True)
class DOM(Template):
    """Template for generating a tree view of DOM etree objects."""

    items_lookup: Callable[[Any], Any] = getattr_maker('childNodes')
    type_name_lookup: Callable[[Any], str] = getattr_maker('nodeName')
    value_lookup: Callable[[Any], Any] = getattr_maker('text')


@dataclasses.dataclass(slots=True, frozen=True)
class HTML(Template):
    """Template for generating a tree view of HTML etree objects."""

    type_name_lookup: Callable[[Any], str] = getattr_maker('tag')
    value_lookup: Callable[[Any], Any] = getattr_maker('text')


@dataclasses.dataclass(slots=True, frozen=True)
class XML(Template):
    """Template for generating a tree view of XML etree objects."""

    type_name_lookup: Callable[[Any], str] = getattr_maker('tag')
    value_lookup: Callable[[Any], Any] = getattr_maker('text')


class _MaxSearchError(Exception):
    """Reached maximum number of nodes to be searched."""

    pass


class _NodeInfo:
    """Non-recursive tree node info."""

    def __init__(self, obj: Any, node_key: _NodeKey, config: Template,
                 nodes_searched: int, ancestors_ids: set[int], depth: int):

        self.key: _NodeKey = node_key
        self.path: str = node_key.path
        self.config: Template = config
        self.nodes_searched: int = nodes_searched
        self.maxed_search: bool = False
        # For displaying an ellipsis indicating possible inner content
        self.maxed_depth: bool = depth >= config.max_depth
        is_infinite_recursion: bool = id(obj) in ancestors_ids

        self.type_name: str
        try:
            self.type_name = self.config.type_name_lookup(obj)
        except AttributeError:
            self.type_name = '?'
        original_var: Any = obj
        obj = self.config.items_lookup(obj)

        # These refer to the contents of Maps or Sequences
        self.items_key_type: KeyType = KeyType.NONE
        self.items_len: int | None = None
        self.update_items_info(obj, original_var)

        self.branches: dict[_NodeKey, Any] = {}
        if not is_infinite_recursion:
            try:
                self.add_branches(obj)
            except _MaxSearchError:
                pass

        self.var_repr: str = f'<{self.type_name}>'
        if is_infinite_recursion:
            self.var_repr = f'<...> {self.var_repr}'
        if self.config.show_lengths and self.items_len is not None:
            self.var_repr = f'{self.var_repr}[{self.items_len}]'

    def update_items_info(self, var: Any, original_var: Any):
        """Check which kind of iterable var is, if any, and update.

        Update `self.items_key_type` and `self.items_len` if needed.
        """
        # Check for Collections
        try:
            var, size = self.get_var_len(var, original_var)
        # If TypeError, var is not a Collection of items
        # Also ignore possible attribute lookup error
        except (TypeError, AttributeError):
            return

        # Check for Mappings
        if self.is_mapping(var):
            self.items_key_type = KeyType.MAP
            self.items_len = size
            return

        # Check for Sequences
        if self.is_sequence(var):
            self.items_key_type = KeyType.INDEX
            self.items_len = size
            return

        # Check for set-like Collections
        if self.is_collection(var):
            self.items_key_type = KeyType.SET
            self.items_len = size

    def get_var_len(self, var: Any, original_var: Any) -> tuple[Any, int]:
        """Check for contents in `var`.

        If no content is found, use the value lookup for tree leaves.
        Return the value and the number of items it contains.
        Raise :exc:`TypeError` if not a Collection.
        """
        size: int = len(var)  # Raises TypeError if not a Collection
        if not size:
            var = self.config.value_lookup(original_var)
            # value_lookup might return itself and var might be empty
            size = len(var)
        if isinstance(var, str | bytes | bytearray):
            raise TypeError
        return var, size

    @staticmethod
    def is_mapping(var: Any) -> bool:
        try:
            # Since Maps are usually also Sequences, the priority is
            # to access the Map items. But if the keys do not match
            # their Sequence values, the priority inverts.
            if (all(key1 == key2 for key1, key2 in zip(var, var.keys()))
                    and all(var[key] == value for key, value in var.items())
                    and len(var) == len(var.keys()) == len(var.items())):
                return True
        except (AttributeError, TypeError, KeyError):
            pass
        return False

    @staticmethod
    def is_sequence(var: Any) -> bool:
        try:
            if all(var[index] == value for index, value in enumerate(var)):
                return True
        except (KeyError, TypeError):
            pass
        return False

    @staticmethod
    def is_collection(var: Any) -> bool:
        try:
            next(iter(var))
        # Unknown type: has len, but is not an iterable. Ignore contents.
        except (TypeError, KeyError):
            return False
        # Is an empty Collection
        except StopIteration:
            pass
        return True

    def add_branches(self, var: Any):
        if self.config.include_attributes and hasattr(var, '__dict__'):
            for key, value in vars(var).items():
                if self.include_attr(key):
                    self.add_branch(KeyType.ATTR, key, value)
        if self.config.include_dir:
            for key in dir(var):
                if self.include_attr(key):
                    try:
                        value = getattr(var, key)
                    except AttributeError:
                        continue
                    self.add_branch(KeyType.ATTR, key, value)
        match self.items_key_type:
            case KeyType.MAP:
                for key, value in var.items():
                    self.add_branch(KeyType.MAP, key, value)
            case KeyType.INDEX:
                for index, value in enumerate(var):
                    self.add_branch(KeyType.INDEX, index, value)
            case KeyType.SET:
                for value in var:
                    self.add_branch(KeyType.SET, 1, value)
        # Success -- do not display ellipsis indicating max depth exceeded
        self.maxed_depth = False

    def add_branch(self, key_type: KeyType, key: Any, value: Any):
        if self.nodes_searched >= self.config.max_search:
            self.maxed_search = True
            if not self.branches:  # empty
                # For displaying an ellipsis indicating possible inner content
                self.maxed_depth = True
            raise _MaxSearchError
        if self.maxed_depth:
            raise _MaxSearchError
        node_key = _NodeKey(key_type, key)
        self.branches[node_key] = value
        self.nodes_searched += 1

    def include_attr(self, key: str) -> bool:
        if key.startswith('__') and key.endswith('__'):
            return self.config.include_special
        if key.startswith('_'):
            return self.config.include_protected
        return True

    def __lt__(self, other):
        if self.key.type == other.key.type == KeyType.SET:
            if self.key.counter != other.key.counter:
                return self.key.counter < other.key.counter
            return self.path < other.path
        return self.key < other.key

    def __str__(self) -> str:
        return f'{self.key}{self.var_repr}'

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self})'


class _InfoTree:
    """A recursive tree builder of :class:`_NodeInfo`."""

    def __init__(self, obj: Any, node_key: _NodeKey, config: Template,
                 nodes_searched: int, ancestors_ids: set[int], depth: int = 0):
        self.config: Template = config
        self.ancestors_ids: set[int] = ancestors_ids.copy()
        self.depth: int = depth
        self.node_info: _NodeInfo = _NodeInfo(
            obj, node_key, config, nodes_searched, ancestors_ids, self.depth
        )
        self.nodes_searched: int = self.node_info.nodes_searched
        self.ancestors_ids.add(id(obj))
        self.is_complete: bool = not self.node_info.branches
        self.maxed_depth: bool = (bool(self.depth >= config.max_depth
                                       and self.node_info.branches)
                                  or self.node_info.maxed_depth)
        self.has_maxed_depth: bool = self.maxed_depth
        self.branches: list[_InfoTree] = []
        self.updated: bool = False

    def update(self):
        """Recursively call to update the deepest :class:`_InfoTree`."""
        if self.is_complete:
            return
        elif self.updated:
            for info_tree in self.branches:
                info_tree.nodes_searched = self.nodes_searched
                info_tree.update()
                self._update_info_tree(info_tree)
        else:
            self.updated = True
            for key, var in self.node_info.branches.items():
                info_tree: _InfoTree = _InfoTree(
                    var, key, self.config, self.nodes_searched,
                    self.ancestors_ids, self.depth + 1,
                )
                self.branches.append(info_tree)
                self._update_info_tree(info_tree)
        self.is_complete = all(info_tree.is_complete
                               for info_tree in self.branches)

    def _update_info_tree(self, info_tree: '_InfoTree'):
        if info_tree.has_maxed_depth:
            self.has_maxed_depth = True
        if self.nodes_searched == info_tree.nodes_searched:
            info_tree.is_complete = True
        self.nodes_searched = info_tree.nodes_searched


class _SubtreeCreator:
    """Create a :class:`Subtree` instance.

    Used for pre-computing and analysing the branches.
    Needed because Subtree inherits from tuple, which is immutable, and
    the creation process is too complex to be done inside
    :meth:`_SubtreeCreator.__new__`.
    """

    def __init__(self, cls: Type['Subtree'], info_tree: _InfoTree):
        self.config: Template = info_tree.config
        info: _NodeInfo = info_tree.node_info
        node_key: _NodeKey = info.key
        self.all_branches: tuple[Subtree, ...] = tuple(
            Subtree(sub_info_tree) for sub_info_tree in info_tree.branches
        )
        self.group_branches()
        max_branches: int = int(min(float(len(self.all_branches)),
                                    self.config.max_branches))
        overflowed: bool = info.maxed_search or (len(self.all_branches)
                                                 > self.config.max_branches)
        self.subtree: Subtree = tuple.__new__(
            cls, self.all_branches[:max_branches]
        )
        self.subtree._node_key = node_key
        self.subtree._config = self.config
        self.subtree._info = info
        self.subtree._label = str(info)
        self.subtree._overflowed = overflowed
        self.subtree._maxed_depth = info.maxed_depth

    @staticmethod
    def group_to_map(v: list[set[int]]) -> dict[tuple[int, int], int]:
        """Convert groups of index mappings to consecutive range maps.

        Argument `v` is a list of indices grouped in sets that map to
        the same structure in a Sequence tree. Their positions indicate
        where they map to. Example:
            `branches = [A, B, A, A, C, A, B]`
            `unique_branches = [A, B, C]`
            `v = [{0, 2, 3, 5}, {1, 6}, {4}]`
        means that `A` shows in indices `v[0] = {0, 2, 3, 5}`, `B` shows
        in indices `v[1] = {1, 6}`, and `C` shows in `v[2] = {4}`.

        The return value is a dict of sequential ranges of indices
        as keys and their mapping to unique_branches. In the previous
        case it will return `{(0, 1): 0, (1, 2): 1, (2, 4): 0,
        (4, 5): 2, (5, 6): 0, (6, 7): 1}`.
        """
        if not v:
            return {}
        u: dict[tuple[int, int], int] = {}
        for k, s in enumerate(v):
            if not s:
                continue
            sv: list[int] = list(sorted(s))
            su: list[tuple[int, int]] = [(sv[0], sv[0] + 1)]
            for x in sv[1:]:  # type: int
                if x == su[-1][1]:
                    su[-1] = (su[-1][0], x + 1)
                else:
                    su.append((x, x + 1))
            for t in su:  # type: tuple[int, int]
                u[t] = k
        # noinspection PyTypeChecker
        return dict(sorted(u.items()))

    # noinspection PyProtectedMember
    def group_branches(self):
        """Group equivalent branches (with the same type structure)."""
        if not self.all_branches:
            return
        # Group unique consecutive Sequence branches and
        # unique Collection branches
        unique_index_branches: list[Subtree] = []
        index_branch_groups: list[list[Subtree]] = []
        index_key_groups: list[set[int]] = []
        unique_set_branches: list[Subtree] = []
        added_branches: list[Subtree] = []
        for branch in self.all_branches:
            # Process Sequences
            if branch._node_key.type == KeyType.INDEX:
                range_key = range(*branch._node_key.slice)
                try:
                    index = unique_index_branches.index(branch)
                except ValueError:
                    unique_index_branches.append(branch)
                    index_branch_groups.append([branch])
                    index_key_groups.append(set(range_key))
                else:
                    index_branch_groups[index].append(branch)
                    index_key_groups[index].update(range_key)
            # Process Collections
            elif branch._node_key.type == KeyType.SET:
                try:
                    index = unique_set_branches.index(branch)
                except ValueError:
                    branch._node_key.reset_counter()
                    unique_set_branches.append(branch)
                else:
                    unique_set_branches[index]._node_key.increment_counter()
            # Do not group other types
            else:
                added_branches.append(branch)

        # Update node keys
        unique_index_branches = self.update_index_keys(index_branch_groups,
                                                       index_key_groups)
        for branch in unique_set_branches:
            branch._update_key(branch._node_key)

        # Join and sort
        self.all_branches = tuple(
            added_branches + unique_index_branches + unique_set_branches
        )
        self.sort_branches()

    # noinspection PyProtectedMember
    def update_index_keys(self, branch_groups: list[list['Subtree']],
                          key_groups: list[set[int]]) -> list['Subtree']:
        updated_branches: list[Subtree] = []
        for _range, index in self.group_to_map(key_groups).items():
            branch = branch_groups[index].pop()
            if self.config.show_lengths:
                branch._update_key(_NodeKey(KeyType.INDEX, _range))
            else:
                branch._update_key(_NodeKey(KeyType.INDEX, None))
            updated_branches.append(branch)
        return list(sorted(updated_branches, key=lambda x: x.key))

    # noinspection PyProtectedMember
    def sort_branches(self):
        self.all_branches = tuple(sorted(
            self.all_branches,
            key=((lambda x: x._info) if self.config.sort_keys
                 else lambda x: x._node_key._type)
        ))


@functools.total_ordering
class Subtree(tuple):
    """A recursive object tree structure."""

    # Initialized by _SubtreeCreator
    _node_key: _NodeKey
    _config: Template
    _info: _NodeInfo
    _label: str
    _overflowed: bool
    _maxed_depth: bool
    # Initialized by __init__
    _nodes: int
    _hash: int
    _path: str
    _is_expandable: bool

    def __new__(cls, info_tree: _InfoTree):
        """Create an immutable nested subtree."""
        return _SubtreeCreator(cls, info_tree).subtree

    def __init__(self, *_args, **_kwargs):
        """Initialize the remaining subtree attributes."""
        self._update_paths()
        node: Subtree
        self._nodes = len(self) + sum(node._nodes for node in self)
        # Hash is unique if overflowed or max depth reached because
        # the tree is incomplete and equality between incomplete trees
        # cannot be established.
        self._hash = hash((
            self._info.var_repr,
            self._info.key.type,
            id(self)*(self._overflowed or self._maxed_depth),
            tuple(map(hash, self)),
        ))
        self._path = self._node_key.path
        self._update_paths()
        self._is_expandable = bool(self or self._overflowed
                                   or self._maxed_depth)

    @property
    def key(self) -> str:
        """Key path from previous node to the current one."""
        return self._node_key.path

    @property
    def key_type(self) -> KeyType:
        """Return the key type used.

        Can be a Sequence index, a Mapping key, an attribute, a Set
        (empty key), or none (also empty).
        """
        return self._node_key.type

    @property
    def path(self) -> str:
        """Return the key path from the root node to the current one."""
        return self._path

    @property
    def type(self) -> str:
        """Return the type name of the object the node corresponds to."""
        return self._info.type_name

    @property
    def label(self) -> str:
        """Return the full text displayed for the node."""
        return self._label

    @property
    def nodes(self) -> int:
        """Return the number of inner nodes indexed."""
        return self._nodes

    @property
    def config(self) -> Template:
        """Return the settings used for generating the tree."""
        return self._config

    @property
    def is_expandable(self) -> bool:
        """Return a flag for whether the node has inner content.

        If true, the node will show as expandable in the tree view.
        """
        return self._is_expandable

    @property
    def maxed_depth(self) -> bool:
        """Return a flag for maximum depth reached.

        Only evaluate to `True` if the object the node refers to has
        inner content that was not indexed.
        """
        return self._maxed_depth

    @property
    def overflowed(self) -> bool:
        """Return a flag for maximum branch exceeded.

        If `True`, the node has more branches than indexed.
        Might be from reaching either `max_search` or `max_branches`.
        """
        return self._overflowed

    def _update_key(self, new_key: _NodeKey):
        self._node_key = new_key
        self._info.key = new_key
        self._label = str(self._info)
        self._update_paths()

    def _update_paths(self, parent_path: str = ''):
        if self._node_key.type == KeyType.SET:
            self._path = f'{parent_path}.copy().pop()'
        else:
            self._path = f'{parent_path}{self._node_key.path}'
        for branch in self:  # type: Subtree
            branch._update_paths(self._path)

    def _get_tree_lines(self, max_lines: float,
                        root_pad: str = '',
                        branch_pad: str = '') -> list[str]:
        lines: list[str] = [f'{root_pad}{self._label}']
        if self._maxed_depth and self._is_expandable:
            lines.append(f'{branch_pad}└── ...')
            return lines
        if not self:  # empty
            return lines

        end: int = len(self)
        if not self._overflowed:
            end -= 1
        for branch in self[:end]:  # type: Subtree
            lines.extend(branch._get_tree_lines(
                max_lines=max_lines,
                root_pad=f'{branch_pad}├── ',
                branch_pad=f'{branch_pad}│   ',
            ))
        if self._overflowed:
            lines.append(f'{branch_pad}...')
        else:
            last_branch: Subtree = self[-1]
            lines.extend(last_branch._get_tree_lines(
                max_lines=max_lines,
                root_pad=f'{branch_pad}└── ',
                branch_pad=f'{branch_pad}    ',
            ))
        if len(lines) > max_lines:
            del lines[int(max_lines) - 1:]
            lines.append(' ...')
        return lines

    def to_dict(self, max_lines: float) -> str | dict[str, str | dict]:
        """Create a nested dict representing the type tree structure."""
        if not self:  # empty
            return str(self._info)
        branch: Subtree
        return {
            str(branch._info): branch.to_dict(max_lines)
            for branch in self
        }

    def __eq__(self, other: object) -> bool:
        """Compare two subtrees.

        Subtrees containing the same type structure are equivalent,
        except for set-like nodes.
        """
        if not isinstance(other, type(self)):
            raise NotImplementedError
        if self._hash != other._hash:
            return False
        if self._info.var_repr != other._info.var_repr:
            return False
        if self._info.key.type != other._info.key.type:
            return False
        if self._overflowed or self._maxed_depth:
            return False
        return super().__eq__(other)

    def __lt__(self, other: object) -> bool:
        """Compare two subtrees for sorting."""
        if not isinstance(other, type(self)):
            raise NotImplementedError
        if self == other:
            return False
        if self._info.key.type != other._info.key.type:
            if self._info.key.type is None:
                return True
            if other._info.key.type is None:
                return False
            return self._info.key.type < other._info.key.type
        if self._info.var_repr != other._info.var_repr:
            return self._info.var_repr < other._info.var_repr
        if len(self) != len(other):
            return len(self) < len(other)
        for x, y in zip(self, other):
            if x != y:
                return x < y
        return self._hash < other._hash

    def __str__(self) -> str:
        """Return the string representation of the subtree."""
        return f'{self._info!s}{{...}}'

    def __repr__(self) -> str:
        """Return the repr of the subtree."""
        return f'{type(self).__name__}({self!s})'

    def __hash__(self) -> int:
        """Return the hash of the subtree.

        Equivalent (equal) subtrees always evaluate to the same hash.
        """
        return self._hash


class Tree(Subtree):
    """Root node of the type tree representation of a Python object.

    Root-only (:class:`Tree`) attributes:

    :param depth: Maximum depth reached
    :type depth: int
    :param searches: Total number of nodes searches
    :type searches: int

    Inherited (:class:`Subtree`) attributes:

    :param key: Key path from previous node to the current one
    :type key: str
    :param key_type: The key type used. Type `help(KeyType)` for more
        info
    :type key_type: KeyType
    :param path: The key path from the root node to the current one
    :type path: str
    :param type: The type name of the object the node corresponds to
    :type type: str
    :param label: The full text displayed for the node
    :type label: str
    :param nodes: The number of inner nodes indexed
    :type nodes: int
    :param config: The settings used for generating the tree
    :type config: Template
    :param is_expandable: Flag for whether the node has inner content.
        If `True`, the node will show as expandable in the tree view
    :type is_expandable: bool
    :param maxed_depth: Flag for maximum depth reached. If `True`, the
        object the node refers to has inner content that was not indexed
    :type maxed_depth: bool
    :param overflowed: Flag indicating maximum branch exceeded. If
        `True`, the node has more branches than indexed. Might be from
        reaching either `max_search` or `max_branches`
    :type overflowed: bool
    """

    _depth: int
    _searches: int
    max_lines: float

    def __new__(cls, obj: Any, *, key_text: str | None = None,
                max_lines: float = _DEFAULT_MAX_LINES,
                template: Type[Template] = Template, **kwargs):
        """Use breadth-first search to construct a new instance."""
        # Since Tree inherits from tuple, which is immutable, the
        # instance size must be pre-computed before the instance
        # initialization. This process is done recursively with a
        # breadth-first algorithm.

        # noinspection PyArgumentList
        config: Template = template(**kwargs)
        info_tree: _InfoTree = _InfoTree(
            obj, _NodeKey(KeyType.NONE, key_text), config,
            nodes_searched=0, ancestors_ids=set(), depth=0
        )
        depth: int = 0
        while not info_tree.is_complete:
            info_tree.update()
            depth += 1
        tree: Tree = super().__new__(cls, info_tree)
        tree._depth = depth
        tree._searches = info_tree.nodes_searched
        tree.max_lines = max_lines
        return tree

    def __init__(self, obj: Any, **kwargs):
        """Build a recursive object tree structure.

        :param obj: Any Python object to be analysed
        :type obj: Any
        :param key_text: Placeholder text for the root key node.
            Defaults to None
        :type key_text: str, optional
        :param template: A configuration template for common object
            types. Currently supported: `Template` (default), `DOM`,
            `HTML`, and `XML`
        :param items_lookup: Function used to access the node's content.
            Defaults to `lambda var: var`
        :type items_lookup: Callable[[Any], Any], optional
        :param type_name_lookup: Function used to get the type name.
            Defaults to `lambda var: type(var).__name__`
        :type type_name_lookup: Callable[[Any], Any], optional
        :param value_lookup: Function used to get the value when the
            node's content is empty (tree leaves). Defaults to
            `lambda var: var`
        :type value_lookup: Callable[[Any], Any], optional
        :param sort_keys: Flag for sorting keys alphabetically. Defaults
            to `True`
        :type sort_keys: bool, optional
        :param show_lengths: Flag for displaying sizes of iterables.
            This affects how subtrees are grouped together, since
            Sequences of different sizes but same content types are
            considered equivalent. Defaults to `True`
        :type show_lengths: bool, optional
        :param include_attributes: Flag for including the mutable
            attributes returned by `vars()`. Defaults to `True`
        :type include_attributes: bool, optional
        :param include_dir: Flag for including the attributes returned
            by `dir()`, except the protected (`_protected`) and special
            (`__special__`) ones. Defaults to `False`
        :type include_dir: bool, optional
        :param include_protected: Flag for including the protected
            (`_protected`) attributes.  Defaults to `False`
        :type include_protected: bool, optional
        :param include_special`: Flag for including the special
            (`__special__`) attributes.  Defaults to `False`
        :type include_special: bool, optional
        :param max_lines: Maximum number of lines to be printed For the
            GUI, it is the maximum number of rows to be displayed, not
            including the extra ellipsis at the end. Can be disabled by
            setting it to infinity (`float('inf')` or `math.inf`).
            Defaults to 1000
        :type max_lines: float, optional
        :param max_search: Maximum number of nodes searched. Defaults to
            100,000
        :type max_search: float, optional
        :param max_depth: Maximum search depth. Defaults to 20
        :type max_depth: float, optional
        :param max_branches: Maximum number of branches displayed on
            each node. This only applies after grouping. Defaults to
            infinity
        :type max_branches: float, optional
        """
        super().__init__(obj, **kwargs)

    @property
    def depth(self) -> int:
        """Maximum depth reached."""
        return self._depth

    @property
    def searches(self) -> int:
        """Total number of nodes searched."""
        return self._searches

    def to_dict(self, max_lines: float | None = None
                ) -> str | dict[str, str | dict]:
        """Return a nested `dict` representation of the type tree."""
        if max_lines is None:
            max_lines = self.max_lines
        return super().to_dict(max_lines)

    def to_json(self, *args, max_lines: float | None = None, **kwargs) -> str:
        """Return a JSON representation of the type tree."""
        if max_lines is None:
            max_lines = self.max_lines
        return json.dumps(self.to_dict(max_lines), *args, **kwargs)

    def save_as_json(self, file_path: str, *args,
                     max_lines: float | None = None,
                     encoding: str = 'utf-8',
                     ensure_ascii: bool = False,
                     indent: int = 4,
                     **kwargs):
        """Save a JSON representation of the type tree to a file."""
        if max_lines is None:
            max_lines = self.max_lines
        with open(file_path, 'w', encoding=encoding) as file:
            json.dump(self.to_dict(max_lines), file, *args,
                      ensure_ascii=ensure_ascii,
                      indent=indent,
                      **kwargs)

    def save_as_text(self, file_path: str, encoding: str = 'utf-8', **kwargs):
        """Save the tree view as a text file."""
        with open(file_path, 'w', encoding=encoding) as f:
            f.write(self.to_string(**kwargs))

    def to_string(self, max_lines: float | None = None,
                  verbose: bool = False) -> str:
        """Get a tree view of the object's type structure as a string.

        :param max_lines: Maximum number of lines to be printed. Can be
            disabled by setting it to infinity
        :type max_lines: float, optional
        :param verbose: Flag for printing extra information before
            printing the tree view. Defaults to False
        :type verbose: bool, optional
        """
        lines: list[str] = []
        if verbose:
            lines.append('Nodes (indexed/searched): '
                         f'{self.nodes}/{self.searches}')
            lines.append(f'Maximum depth: {self.depth}')
        if max_lines is None:
            max_lines = self.max_lines
        lines.extend(self._get_tree_lines(max_lines, ' ', ' '))
        return '\n'.join(lines)

    def print(self, max_lines: float | None = None, verbose: bool = False):
        """Print a tree view of the object's type structure.

        :param max_lines: Maximum number of lines to be printed. Can be
            disabled by setting it to infinity
        :type max_lines: float, optional
        :param verbose: Flag for printing extra information before
            printing the tree view. Defaults to False
        :type verbose: bool, optional
        """
        print(self.to_string(max_lines=max_lines, verbose=verbose))

    def view(self, spawn_thread: bool = True, spawn_process: bool = False,
             max_lines: float | None = None):
        """Show a tree view of the object's type structure in a GUI.

        :param spawn_thread: Run the GUI in a separate thread
        :type spawn_thread: bool
        :param spawn_process: Run the GUI in a separate process
        :type spawn_process: bool
        :param max_lines: Maximum number of rows to be displayed, not
            including the extra ellipsis at the end. Can be disabled by
            setting it to infinity
        """
        if max_lines is None:
            max_lines = self.max_lines
        tree_viewer(self, max_lines,
                    spawn_thread=spawn_thread,
                    spawn_process=spawn_process)


def print_tree(obj: Any, *, max_lines: float | None = None,
               verbose: bool = False, **kwargs):
    """Print a tree view of the object's type structure.

    :param obj: Any Python object to be analysed
    :type obj: Any
    :param max_lines: Maximum number of lines to be printed. Can be
        disabled by setting it to infinity
    :type max_lines: float, optional
    :param verbose: Flag for printing extra information before printing
        the tree view. Defaults to False
    :type verbose: bool, optional
    :param kwargs: Same as :class:`Tree`. Type `help(Tree.__init__)` for
        the full list
    """
    Tree(obj, **kwargs).print(max_lines=max_lines, verbose=verbose)


def view_tree(obj: Any, *,
              spawn_thread: bool = True,
              spawn_process: bool = False,
              **kwargs):
    """Show a tree view of the object's type structure in a GUI.

    :param obj: Any Python object to be analysed
    :type obj: Any
    :param spawn_thread: Run the GUI in a separate thread
    :type spawn_thread: bool
    :param spawn_process: Run the GUI in a separate process
    :type spawn_process: bool
    :param kwargs: Same as :class:`Tree`. Type `help(Tree.__init__)` for
        the full list
    """
    Tree(obj, **kwargs).view(spawn_thread=spawn_thread,
                             spawn_process=spawn_process)



"""Module for loading .ico files."""

import os

ICON_DIR = 'icons'

# Look for the icons subdirectory in the same directory as this module
try:
    _icon_dir = os.path.join(os.path.dirname(__file__), ICON_DIR)
except NameError:
    _icon_dir = ICON_DIR
if os.path.isdir(_icon_dir):
    ICON_DIR = _icon_dir
elif not os.path.isdir(ICON_DIR):
    raise RuntimeError('Icon directory not found: '
                       f'"{os.path.normpath(ICON_DIR)}"')


class FileError(Exception):
    """Exception class for file-related errors."""

    pass


def fix_extension(file_path: str, ext: str) -> str:
    """Fix missing extensions."""
    folder, file_name = os.path.split(file_path)
    if not file_name:
        raise FileError('File name is required')
    file_root, file_ext = os.path.splitext(file_name)
    if file_ext != ext:
        return file_path + ext
    return file_path


def load_ico(file_name: str, size: float | None = None) -> bytes | list[bytes]:
    """Load a .ico file.

    If no size is given, return all images in the ICO file. Otherwise,
    return the image in the ICO file that closest match the given size.
    """
    ico_path = os.path.join(ICON_DIR, fix_extension(file_name, '.ico'))
    with open(ico_path, 'rb') as f:
        b = f.read()

    pos = 4
    n_images = int(b[pos])
    offset = 6 + 16*n_images
    file_sizes = []
    best_match_index = 0
    best_match_error = float('inf')
    pos += 2
    n = 4
    for index in range(n_images):
        width = int(b[pos])
        height = int(b[pos + 1])

        pos += 8
        file_sizes.append(int.from_bytes(b[pos:pos + n], 'little'))

        pos += n
        if offset != int.from_bytes(b[pos:pos + n], 'little'):
            raise FileError('Image file size not consistent with byte offset')
        offset += file_sizes[-1]
        pos += n

        if size is None:
            continue
        error = abs((width - size)*(height - size))
        if error < best_match_error:
            best_match_index = index
            best_match_error = error

    if size is not None:
        pos += sum(file_sizes[:best_match_index])
        return b[pos:pos + file_sizes[best_match_index]]

    images = []
    for file_size in file_sizes:
        images.append(b[pos:pos + file_size])
        pos += file_size
    return images



"""Module for handling the GUI."""

import math
import multiprocessing
import threading
import tkinter as tk
from typing import Any, Optional

try:
    from ico import load_ico
except (ModuleNotFoundError, ImportError):
    from .ico import load_ico

# Make text look sharper for displays with scaling different from 100%.
# Only works for Windows.
# WARNING: This is permanently applied to the whole process, not just
# the window or the thread.
SET_DPI_AWARENESS: bool = True
try:
    # This is for calling windll.shcore.SetProcessDpiAwareness(1) on
    # the initialization of a new ViewTreeWindow object.
    from ctypes import windll  # type: ignore
except ImportError:
    pass  # Not required

_PYPERCLIP_LOADED: bool = True
try:
    import pyperclip
except ModuleNotFoundError:
    _PYPERCLIP_LOADED = False  # Not required

REFERENCE_DPI: float = 96


# This always returns the same value, so it can only be called after
# setting the DPI awareness.
def get_dpi() -> float:
    """Get monitor DPI."""
    screen = tk.Tk()
    current_dpi: float = screen.winfo_fpixels('1i')
    screen.destroy()
    return current_dpi


class TreeNode:
    """Recursive class representing each tree node."""

    font_size: int = 10
    # The values below will be multiplied by font_size
    row_height_factor: float = 2.1
    indent_width_factor: float = 2.3
    text_pad_factor: float = 0.2
    icon_size_factor: float = 1.0

    line_style: dict[str, Any] = {'fill': 'gray60'}  # , 'dash': (1, 1)}
    label_style: dict[str, dict[str, Any]] = {
        'normal': {
            'foreground': 'black',
            'background': 'gray97',
        },
        'selected': {
            'foreground': 'black',
            'background': 'gray60',
        },
    }
    icon_files: dict[str, bytes] = {}  # Class cache for icon files

    row_height: int = round(font_size*row_height_factor)
    indent_width: int = round(font_size*indent_width_factor)
    text_pad: int = round(font_size*text_pad_factor)
    # Not true icon size. It picks the image in the .ico files with
    # dimensions closest to the one given by icon_size.
    icon_size: float = font_size*icon_size_factor

    @classmethod
    def update_dpi(cls):
        """Fix sizes according the DPI scaling of the monitor."""
        size = cls.font_size*get_dpi()/REFERENCE_DPI
        cls.row_height = round(size*cls.row_height_factor)
        cls.indent_width = round(size*cls.indent_width_factor)
        cls.text_pad = round(size*cls.text_pad_factor)
        cls.icon_size = size*cls.icon_size_factor

    def __init__(self, root: 'ViewTreeWindow', parent: Optional['TreeNode'],
                 tree: 'PicklableTree'):
        """Recursively initialize the tree, but wait for drawing later."""
        self.font: tuple[str, int] = ('Consolas', self.font_size)
        # Cache of PhotoImage instances for icons
        self.icons: dict[str, tk.PhotoImage] = {}
        self.root: ViewTreeWindow = root
        self.canvas: tk.Canvas = root.canvas
        self.statusbar: tk.Label = root.statusbar
        self.parent: TreeNode | None = parent
        self.tree: PicklableTree = tree
        self.is_expanded: bool = True
        self.is_visible: bool = True
        self.is_selected: bool = False
        self.children: list[TreeNode] = []
        self.x: int | None = None
        self.y: int | None = None
        self.label = None
        self.index: int = len(self.root.all_nodes)

        self.root.n_lines += 1
        self.lines_maxed: bool = False
        self.root.all_nodes.append(self)
        if self.tree.overflowed or self.tree.maxed_depth:
            self.root.n_lines += 1
        for branch in tree:
            if self.root.n_lines >= self.tree.max_lines:
                self.lines_maxed = True
                break
            self.children.append(TreeNode(self.root, self, branch))

    def refresh(self):
        """Refresh the whole canvas."""
        if self.parent:
            self.parent.refresh()
        else:
            self.canvas['cursor'] = 'watch'
            self.canvas.update()
            self.canvas.delete(tk.ALL)
            self.draw()
            self.root.on_resize()
            # Cursor should update to 'hand2' if over a node icon
            if self.canvas['cursor'] == 'watch':
                self.canvas['cursor'] = 'arrow'

    def update_cursor(self, event):
        """Update the mouse cursor."""
        if event.type == tk.EventType.Enter:
            self.canvas['cursor'] = 'hand2'
        elif event.type == tk.EventType.Leave:
            self.canvas['cursor'] = 'arrow'

    def update_visibility(self):
        """Update the .is_visible flags."""
        if self.is_visible and self.is_expanded:
            for child in self.children:
                child.is_visible = True
                child.update_visibility()
        else:
            for child in self.children:
                child.is_visible = False
                child.update_visibility()

    def draw(self, x=0, y=0) -> int:
        """Draw a node and recursively call itself for each child node.

        Return the next available vertical position for drawing.
        """
        dy = self.row_height//2
        icon_width = self.draw_icon(x + dy, y + dy)
        self.x = x + (self.row_height + icon_width)//2 + self.text_pad
        self.y = y
        self.draw_text()
        x2 = x + self.indent_width
        y2 = y + self.row_height

        if (not self.children and not self.lines_maxed
                and not self.tree.maxed_depth):
            return y2
        if not self.is_expanded:
            return y2

        line_x = x + dy
        line_y = y2 + dy
        last_y = line_y

        if (self.tree.is_expandable and self.tree.maxed_depth
                and not self.tree.overflowed):
            # If max depth reached, draw ellipsis
            last_y = line_y
            # Horizontal line
            self.draw_line(line_x, line_y, self.indent_width, 0)
            self.draw_overflow(x2 + dy + self.text_pad, y2)
            y2 += self.row_height
            line_y = y2 + dy

        for child in self.children:
            last_y = line_y
            # Horizontal line
            self.draw_line(line_x, line_y, self.indent_width, 0)
            y2 = child.draw(x2, y2)
            line_y = y2 + dy

        if self.lines_maxed or self.tree.overflowed:
            # If max lines or max branches reached, draw ellipsis
            self.draw_overflow(line_x - self.text_pad, y2)
            last_y = y2 + self.row_height//4  # Small additional line
            y2 += self.row_height

        # Vertical line
        _id = self.draw_line(line_x, y + dy, 0, last_y - y - dy)
        self.canvas.tag_lower(_id)  # Display under the icons

        return y2

    def draw_line(self, x: int, y: int, dx: int, dy: int) -> int:
        """Draw a line from `(x, y)` to `(x + dx, y + dy)`."""
        return self.canvas.create_line(x, y, x + dx, y + dy, **self.line_style)

    def draw_icon(self, x: int, y: int) -> int:
        """Draw the plus or the minus icon if it is expandable.

        Return the icon width.
        """
        if not self.tree.is_expandable:
            return 0
        if self.is_expanded:
            icon_name = 'minus'
            callback = self.collapse
        else:
            icon_name = 'plus'
            callback = self.expand
        icon = self.get_icon(icon_name)
        _id = self.canvas.create_image(x, y, image=icon)
        self.canvas.tag_bind(_id, '<1>', callback)
        self.canvas.tag_bind(_id, '<3>', self.toggle_children)
        self.canvas.tag_bind(_id, '<Double-1>', callback)
        self.canvas.tag_bind(_id, '<Enter>', self.update_cursor)
        self.canvas.tag_bind(_id, '<Leave>', self.update_cursor)
        x1, y1, x2, y2 = self.canvas.bbox(_id)
        x = self.root.winfo_pointerx()
        y = self.root.winfo_pointery()
        if x1 <= x <= x2 and y1 <= y <= y2:
            self.canvas['cursor'] = 'hand2'
        return icon.width()

    def draw_text(self):
        """Draw the key and the value type of the node."""
        text_x = self.x
        text_y = self.y
        text = self.tree.label or ''
        if self.label is None:
            self.label = tk.Label(self.canvas, text=text, bd=0,
                                  padx=self.text_pad, pady=self.text_pad,
                                  font=self.font)
        if self.is_selected:
            self.label.configure(self.label_style['selected'])
        else:
            self.label.configure(self.label_style['normal'])

        height = self.label.winfo_reqheight()
        text_y += (self.row_height - height)//2
        self.canvas.create_window(text_x, text_y,
                                  anchor='nw', window=self.label)
        self.label.bind('<1>', self.select)
        self.label.bind('<Double-1>', self.copy)

    def draw_overflow(self, x, y):
        """Draw an ellipsis if max lines or max branches reached."""
        self.canvas.create_text(
            x, y, text='...', anchor='nw',
            font=self.font, fill=self.label_style['normal']['foreground']
        )

    def select(self, _event=None,
               update_yview=True, update_last_selected=True):
        """Handle the selection and highlighting of a node."""
        # Selection can switch directly or as a result of expanding or
        # collapsing node. If a selected node is collapsed, it remembers
        # the last selected node so that it can be highlighted after it
        # becomes visible again.
        # The update_yview parameter is set to False on expanding nodes
        # so that it can be called separately with different logic.
        if self.is_selected:
            return
        self.deselect_selected()
        self.is_selected = True
        if update_last_selected:
            self.root.last_selected = self
            path = f' - Path: {self.tree.path}' if self.tree.path else ''
            self.root.statusbar.config(text=f'{self.index}{path}')
        self.root.selected = self
        self.draw_text()
        if update_yview:
            self.update_yview()

    def deselect(self, _event=None):
        """Disable highlighting."""
        if not self.is_selected:
            return
        self.is_selected = False
        self.root.selected = None
        if self.is_visible:
            self.draw_text()

    def deselect_selected(self):
        """Check if any node is selected and deselect it."""
        if self.root.selected is None:
            return
        self.root.selected.deselect()

    def expand(self, _event=None):
        """Expand a node and update the vertical scroll view."""
        if not self.tree.is_expandable:
            self.select()
            return
        if self.is_expanded:
            return
        self.is_expanded = True
        self.update_visibility()
        self.refresh()
        if child := self.has_last_selected():
            child.select(update_yview=False, update_last_selected=False)
        self.update_yview(from_expand=True)

    def collapse(self, _event=None):
        """Collapse a node and update the vertical scroll view."""
        if not self.tree.is_expandable:
            return
        if not self.is_expanded:
            return
        self.is_expanded = False
        self.update_visibility()
        self.refresh()
        # Remember the last selected if it is contained in the
        # collapsed nodes
        if (self.root.last_selected is not None
                and self.root.last_selected.is_descendant(self)):
            self.select(update_last_selected=False)

    def toggle_children(self, _event=None):
        """Toggle the expansion state of all child nodes.

        Use the majority state as the reference and apply the inverse
        state to every child node.
        """
        if not self.tree.is_expandable:
            self.select()
            return
        if not self.children:
            return
        collapse: bool = bool(round(
            sum(
                child.is_expanded for child in self.children
            )/len(self.children)
        ))
        if not self.is_expanded:
            self.is_expanded = True
            self.update_visibility()
            collapse = True
        if collapse:
            next_selected: TreeNode | None = None
            for child in self.children:
                child.is_expanded = False
                child.update_visibility()
                # Remember the last selected if it is contained in the
                # collapsed nodes
                if (next_selected is None
                        and self.root.last_selected is not None
                        and self.root.last_selected.is_descendant(child)):
                    next_selected = child
            self.refresh()
            if next_selected is not None:
                next_selected.select(update_last_selected=False)
        else:  # expand
            last_selected: TreeNode | None = None
            for child in self.children:
                child.is_expanded = True
                child.update_visibility()
                last_selected = last_selected or child.has_last_selected()
            self.refresh()
            if last_selected is not None:
                last_selected.select(update_yview=False,
                                     update_last_selected=False)
            self.update_yview(from_expand=True)

    def update_yview(self, from_expand=False):
        """Update the vertical scroll view."""
        # If it comes from an expanding node, move the canvas up so that
        # it shows as many containing nodes as possible
        top = self.y
        if from_expand:
            bottom = self.visible_children_bottom()
            selected = self.root.selected
            if selected is not None and selected.is_descendant(self):
                selected_bottom = selected.y + self.row_height
            else:
                selected_bottom = self.y + self.row_height
        else:
            bottom = top + self.row_height
            selected_bottom = bottom
        height = bottom - top
        visible_top = self.canvas.canvasy(0)
        visible_height = self.canvas.winfo_height()
        visible_bottom = self.canvas.canvasy(visible_height)
        if visible_top <= top and bottom <= visible_bottom:
            return
        _, y1, _, y2 = map(int, self.canvas['scrollregion'].split())
        canvas_height = y2 - y1
        if top >= visible_top and height <= visible_height:
            rows = (bottom - visible_height)/self.row_height
        elif selected_bottom - top > visible_height:
            rows = (selected_bottom - visible_height)/self.row_height
        else:
            rows = top/self.row_height
        rows = math.ceil(rows)
        self.canvas.yview_moveto(rows*self.row_height/canvas_height)

    def is_descendant(self, other: 'TreeNode') -> bool:
        """Check if `self` is a descendant of `other`."""
        if self.parent == other:
            return True
        if self.parent is None:
            return False
        return self.parent.is_descendant(other)

    def has_last_selected(self):
        """Return the closest visible ancestor if found."""
        if not self.is_selected or not self.is_visible:
            return None
        child = self.root.last_selected
        if child is None or not child.is_descendant(self):
            return None
        while not child.is_visible:
            child = child.parent
        return child

    def visible_children_bottom(self):
        """Bottom y coordinate of visible children."""
        if self.children and self.is_expanded:
            bottom = self.children[-1].visible_children_bottom()
            if self.tree.overflowed:
                return bottom + self.row_height
            return bottom
        elif self.tree.overflowed:
            return self.y + 2*self.row_height
        return self.y + self.row_height

    def get_icon(self, file_name: str) -> tk.PhotoImage:
        """Load the plus/minus icons."""
        if file_name in self.icons:
            return self.icons[file_name]
        try:
            icon_data = self.icon_files[file_name]
        except KeyError:
            icon_data = load_ico(file_name, self.icon_size)
            self.icon_files[file_name] = icon_data
        image = tk.PhotoImage(master=self.canvas, data=icon_data)
        self.icons[file_name] = image
        return image

    def copy(self, _event=None):
        """Copy to clipboard. Called on Ctrl+C or double-click."""
        if not self.is_selected:
            raise Exception('Copy called without being selected')
        if not (path := self.tree.path):
            return 'break'
        if _PYPERCLIP_LOADED:
            # Preferred method
            pyperclip.copy(path)
        else:
            # Tkinter does not really copy to clipboard until a paste
            # occurs before the window is closed.
            self.root.clipboard_clear()
            self.root.clipboard_append(path)
            self.root.update()
        self.root.statusbar.config(text=f'Copied path to clipboard: {path}')
        return 'break'

    def destroy(self):
        """Destroy all content."""
        for c in self.children.copy():
            self.children.remove(c)
            c.destroy()
        self.parent = None


class ScrollbarFrame(tk.Frame):
    """A frame with vertical and horizontal scrollbars.

    The horizontal one automatically hides when it is disabled.
    """

    def __init__(self, master, *, row_height=5, mouse_wheel_rows=3, **opts):
        """Pack the widgets, create bindings, and set focus."""
        self.row_height = row_height
        self.mouse_wheel_rows = mouse_wheel_rows

        super().__init__(master, **opts)

        self.master = master
        self.canvas = tk.Canvas(self, yscrollincrement=row_height, **opts)
        self.vbar = tk.Scrollbar(self, orient="vertical",
                                 command=self.on_vbar)
        self.hbar = tk.Scrollbar(self, orient="horizontal",
                                 command=self.canvas.xview)

        self.canvas.configure(yscrollcommand=self.vbar.set)
        self.canvas.configure(xscrollcommand=self.hbar.set)

        self.vbar.pack(side="right", fill="y")
        self.hbar.pack(side="bottom", fill="x")
        self.is_hbar_packed = True
        self.canvas.pack(side="left", fill="both", expand=True)

        self.frame = tk.Frame(self.canvas)
        self.canvas.create_window((0, 0), window=self.frame, anchor="nw",
                                  tags="self.frame")

        self.master.bind("<Configure>", self.on_resize)
        self.canvas.bind_all("<MouseWheel>", self.on_mouse_wheel)
        self.canvas.bind_all('<Button-4>', self.on_mouse_wheel)
        self.canvas.bind_all('<Button-5>', self.on_mouse_wheel)

        self.canvas.bind('<Key-Prior>', self.page_up)
        self.canvas.bind('<Key-Next>', self.page_down)

        self.on_resize()
        self.canvas.focus_set()

    def on_resize(self, _event=None):
        """Reset the scroll region to encompass the inner frame."""
        _, _, x, y = self.canvas.bbox(tk.ALL)
        w = self.canvas.winfo_width()
        h = self.canvas.winfo_height()

        if w >= x:
            if self.is_hbar_packed:
                h += self.hbar.winfo_height()
                self.toggle_hbar()
        elif not self.is_hbar_packed:
            h -= self.hbar.winfo_height()
            self.toggle_hbar()

        y = self.row_height*math.ceil(y/self.row_height) + 1
        w = max(w, x)
        h = max(h, y)
        self.canvas.itemconfigure('inner', width=w, height=h)
        self.canvas.configure(scrollregion=(0, 0, w, h))

    def toggle_hbar(self):
        """Toggle the horizontal bar visibility state."""
        if self.is_hbar_packed:
            self.hbar.pack_forget()
            self.is_hbar_packed = False
            return
        self.canvas.pack_forget()
        self.hbar.pack(side="bottom", fill="x")
        self.hbar.config(command=self.canvas.xview)
        self.is_hbar_packed = True
        self.canvas.pack(side="left", fill="both", expand=True)

    def page_up(self, _event):
        """Handle page-up key press."""
        self.canvas.yview_scroll(-1, 'pages')
        return 'break'

    def page_down(self, _event):
        """Handle page-down key press."""
        self.canvas.yview_scroll(1, 'pages')
        return 'break'

    def on_vbar(self, *args):
        """Handle vertical bar scrolling logic."""
        _, y1, _, y2 = map(int, self.canvas['scrollregion'].split())
        h = y2 - y1
        y1 = self.canvas.yview()[0]

        match args:
            case 'moveto', str():
                dy = float(args[1]) - y1
            case 'scroll', str(), 'units':
                dy = int(args[1])*self.row_height/h
            case _:
                self.canvas.yview(*args)
                return

        y2 = y1 + self.canvas.winfo_height()/h
        if y1 + dy < 0:
            self.canvas.yview_moveto(0)
        elif y2 + dy > 1:
            d_rows = math.ceil((1 - y2)*h/self.row_height)
            dy = d_rows*self.row_height/h
            self.canvas.yview_moveto(y1 + dy)
        else:
            self.canvas.yview_moveto(y1 + dy)

    def on_mouse_wheel(self, event):
        """Send a call to `self.on_vbar` on mouse wheel movement."""
        sign = 0
        if event.type == tk.EventType.MouseWheel:
            sign = 1 if event.delta < 0 else -1
        elif event.type == tk.EventType.ButtonPress:
            if event.num == 4:
                sign = -1
            elif event.num == 5:
                sign = 1
        units = sign*self.mouse_wheel_rows
        self.on_vbar('scroll', str(units), 'units')


class ViewTreeWindow(tk.Tk):
    """Tkinter GUI with an interactive tree view."""

    width = 540
    height = 720
    background_color = 'gray97'

    def __init__(self, tree):
        """Create a new GUI, set bindings, update icon, and set focus."""
        super().__init__()
        self.title('Tree View')
        self.geometry(f'{self.width}x{self.height}')

        TreeNode.update_dpi()
        self.sf = ScrollbarFrame(self, row_height=TreeNode.row_height,
                                 bg=self.background_color, borderwidth=0,
                                 highlightthickness=0, takefocus=1)
        self.canvas = self.sf.canvas
        self.statusbar = tk.Label(self, text='', bd=1,
                                  relief=tk.SUNKEN, anchor=tk.W)
        self.sf.pack(expand=True, fill='both', side=tk.TOP)
        self.statusbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.sf.on_resize()

        self.selected: TreeNode | None = None
        self.last_selected: TreeNode | None = None
        self.all_nodes: list[TreeNode] = []
        self.n_lines: int = 0
        self.tree: PicklableTree = tree
        self.node: TreeNode = TreeNode(self, None, tree)
        self.node.refresh()
        self.node.expand()
        self.node.select()

        self.bind('<Control-c>', self.copy)
        self.bind('<Key-Up>', self.move_up)
        self.bind('<Key-Down>', self.move_down)
        self.bind('<Key-Left>', self.collapse)
        self.bind('<Key-Right>', self.expand)
        self.bind('<space>', self.toggle_children)

        try:
            images_data = load_ico('icon.ico')
            images = [tk.PhotoImage(data=data) for data in images_data]
            self.iconphoto(False, *images)
        except (FileNotFoundError, tk.TclError):
            pass

        self.focus_force()

    def on_resize(self):
        """Handle screen movement and resizes."""
        self.sf.on_resize()

    def copy(self, _event=None):
        """Copy key path to clipboard."""
        if self.selected is None:
            return 'break'
        self.selected.copy()

    def move_up(self, _event=None):
        """Select the previous node above."""
        if self.selected is None:
            return 'break'
        index = max(self.selected.index, 1)
        for node in self.all_nodes[index - 1::-1]:
            if node.is_visible:
                node.select()
                return

    def move_down(self, _event=None):
        """Select the next node below."""
        if self.selected is None:
            return 'break'
        index = self.selected.index
        for node in self.all_nodes[index + 1:]:
            if node.is_visible:
                node.select()
                break

    def collapse(self, _event=None):
        """Collapse the node."""
        if self.selected is None:
            return 'break'
        if self.selected.is_visible:
            self.selected.collapse()
        else:
            self.move_up()

    def expand(self, _event=None):
        """Expand the node."""
        if self.selected is None:
            return 'break'
        if self.selected.is_visible:
            self.selected.expand()
        else:
            self.move_down()

    def toggle_children(self, _event=None):
        """Toggle the expansion state of child nodes.

        Use majority state and apply the inverse for every child node.
        """
        if self.selected is None or not self.selected.is_visible:
            return 'break'
        self.selected.toggle_children()


class PicklableTree(tuple):
    """Transform Tree into a simpler pickable object."""

    max_lines: float
    label: str
    path: str
    is_expandable: bool
    overflowed: bool
    maxed_depth: bool

    def __new__(cls, tree, max_lines: float):
        """Recursively construct a pickable object for the tree view."""
        self = super().__new__(cls, (
            PicklableTree(subtree, max_lines)  # type: ignore
            for subtree in tree
        ))
        self.max_lines = max_lines
        self.label = tree.label
        self.path = tree.path
        self.is_expandable = tree.is_expandable
        self.overflowed = tree.overflowed
        self.maxed_depth = tree.maxed_depth
        return self


def tree_window_loop(tree: PicklableTree):
    """Start the main loop of the GUI.

    Can be used concurrently, as a separate thread, or as a separate
    process.
    """
    # Make text look sharper for displays with scaling different
    # from 100%. Only works for Windows.
    # WARNING: This is permanently applied to the whole process, not
    # just the window or the thread.
    if SET_DPI_AWARENESS and windll is not None:
        windll.shcore.SetProcessDpiAwareness(1)

    window = ViewTreeWindow(tree)
    window.mainloop()


def tree_viewer(tree, max_lines: float, *,
                spawn_thread=True, spawn_process=False):
    """GUI spawner.

    Convert Tree to a simpler pickable object and optionally spawn a new
    thread or process.
    """
    tree = PicklableTree(tree, max_lines)
    if spawn_process:
        multiprocessing.Process(target=tree_window_loop,
                                args=(tree,)).start()
    elif spawn_thread:
        threading.Thread(target=tree_window_loop, args=(tree,),
                         daemon=True).start()
    else:
        tree_window_loop(tree)
